

import builtins


class ConfigValidator:
    config = None

    def __init__(self, config):
        self.config = config

    def validate(self):
        """Check for all needed parameters in config"""
        if not self.config.index_uid:
            raise ValueError('index_uid is not defined')

        # Start_urls is mandatory
        if not self.config.start_urls and not self.config.sitemap_urls:
            raise ValueError('start_urls is not defined, nor sitemap urls')

        # Start urls must be an array
        if self.config.start_urls and not isinstance(self.config.start_urls,
                                                     list):
            raise builtins.Exception('start_urls should be list')

        # Stop urls must be an array
        if self.config.stop_urls and not isinstance(self.config.stop_urls,
                                                    list):
            raise builtins.Exception('stop_urls should be list')

        # Custom settings must be a dict
        if self.config.custom_settings and not isinstance(self.config.custom_settings,
                                                    dict):
            raise builtins.Exception('custom_settings must be a dictionary')

        if self.config.js_render and not isinstance(self.config.js_render,
                                                    bool):
            raise builtins.Exception('js_render should be boolean')

        # `js_wait` is set to 0s by default unless it is specified
        if self.config.js_wait and not isinstance(self.config.js_wait, int):
            raise builtins.Exception('js_wait should be integer')

        if self.config.use_anchors and not isinstance(self.config.use_anchors,
                                                      bool):
            raise builtins.Exception('use_anchors should be boolean')

        if self.config.sitemap_alternate_links and not isinstance(
                self.config.sitemap_alternate_links, bool):
            raise builtins.Exception('sitemap_alternate_links should be boolean')

        if self.config.sitemap_urls_regexs and not self.config.sitemap_urls:
            raise builtins.Exception(
                'You gave an regex to parse sitemap but you didn\'t provide a sitemap url')

        if self.config.sitemap_urls_regexs and not self.config.sitemap_urls:
            for regex in self.config.sitemap_urls_regex:
                if not isinstance(regex, str):
                    raise builtins.Exception(
                        'You gave an bad regex: ' + regex + ' must be a string')

        if self.config.force_sitemap_urls_crawling and not self.config.sitemap_urls:
            raise builtins.Exception(
                'You want to force the sitemap crawling but you didn\'t provide a sitemap url')

        if not self.config.scrape_start_urls and not self.config.scrap_start_urls:
            raise builtins.Exception(
                'Please use only the new variable name: scrape_start_urls')

        if self.config.nb_hits_max and not isinstance(self.config.nb_hits_max,
                                                      int):
            raise builtins.Exception('nb_hits_max should be integer')



class Hierarchy:
    def __init__(self):
        pass

    @staticmethod
    def get_hierarchy_radio(hierarchy, current_level, levels):
        """Returns the radio hierarchy for the record, where only one level is
        filled and the others are empty
        Ex: {
            lvl0: None,
            lvl1: None,
            lvl2: Baz,
            lvl3: None,
            lvl4: None,
            lvl5: None,
            lvl6: None
        }
        """

        hierarchy_radio = {}
        is_found = False
        for level in reversed(levels):
            if level == 'content':
                continue

            value = hierarchy[level]
            if is_found is False and value is not None and current_level != 'content':
                is_found = True
                hierarchy_radio[level] = value
                continue

            hierarchy_radio[level] = None

        return hierarchy_radio



import builtins
from ..helpers import css_to_xpath


class SelectorsParser:
    @staticmethod
    def _parse_selectors_set(config_selectors):
        selectors_set = {}
        for key in config_selectors:
            if key != 'text':
                selectors_set[key] = config_selectors[key]
            else:
                # Backward compatibility, rename text to content
                key = 'content'
                selectors_set[key] = config_selectors['text']

            # Backward compatibility, if it's a string then we put it in an object
            if isinstance(selectors_set[key], str):
                selectors_set[key] = {'selector': selectors_set[key]}

            # Global
            if 'global' in selectors_set[key]:
                selectors_set[key]['global'] = bool(
                    selectors_set[key]['global'])
            else:
                selectors_set[key]['global'] = False

            # Type
            if 'type' in selectors_set[key]:
                if selectors_set[key]['type'] not in ['xpath', 'css']:
                    raise builtins.Exception(
                        selectors_set[key][
                            'type'] + 'is not a good selector type, it should be `xpath` or `css`')
            else:
                selectors_set[key]['type'] = 'css'

            if selectors_set[key]['type'] == 'css':
                selectors_set[key]['selector'] = css_to_xpath(
                    selectors_set[key]['selector'])

            # We don't need it because everything is xpath now
            selectors_set[key].pop('type')

            # Default value
            selectors_set[key]['default_value'] = selectors_set[key][
                'default_value'] if 'default_value' in \
                                    selectors_set[
                                        key] else None

            # Strip chars
            selectors_set[key]['strip_chars'] = selectors_set[key][
                'strip_chars'] if 'strip_chars' in selectors_set[
                key] else None

            if 'attributes' in selectors_set[key]:
                selectors_set[key][
                    'attributes'] = SelectorsParser._parse_selectors_set(
                    selectors_set[key]['attributes'])

        return selectors_set

    def parse(self, config_selectors):
        selectors = {}

        if 'lvl0' in config_selectors:
            config_selectors = {'default': config_selectors}

        for selectors_key in config_selectors:
            selectors[selectors_key] = self._parse_selectors_set(
                config_selectors[selectors_key])

        return selectors

    @staticmethod
    def parse_min_indexed_level(min_indexed_level):
        min_indexed_level_object = min_indexed_level
        if isinstance(min_indexed_level, int):
            min_indexed_level_object = {
                'default': min_indexed_level_object
            }

        return min_indexed_level_object



"""
DocumentationSpider
"""
from scrapy.linkextractors.lxmlhtml import LxmlLinkExtractor
from scrapy.spiders import CrawlSpider, Rule
from scrapy.http import Request

# Import for the sitemap behavior
from scrapy.spiders import SitemapSpider
from scrapy.spiders.sitemap import regex
import re
import os

# End of import for the sitemap behavior

from scrapy.spidermiddlewares.httperror import HttpError

from scrapy.exceptions import CloseSpider

EXIT_CODE_EXCEEDED_RECORDS = 4


class DocumentationSpider(CrawlSpider, SitemapSpider):
    """
    DocumentationSpider
    """
    http_user = os.environ.get('DOCS_SCRAPER_BASICAUTH_USERNAME', None)
    http_pass = os.environ.get('DOCS_SCRAPER_BASICAUTH_PASSWORD', None)
    meilisearch_helper = None
    strategy = None
    js_render = False
    js_wait = 0
    match_capture_any_scheme = re.compile(r"^(https?)(.*)")
    backreference_any_scheme = r"^https?\2(.*)$"
    # Could be any url prefix such as http://www or http://
    every_schemes = ["http", "https"]
    reason_to_stop = None

    @staticmethod
    def to_any_scheme(url):
        """Return a regex that represent the URL and match any scheme from it"""
        return url if not re.match(
            DocumentationSpider.match_capture_any_scheme, url) else re.sub(
            DocumentationSpider.match_capture_any_scheme,
            DocumentationSpider.backreference_any_scheme, url)

    @staticmethod
    def to_other_scheme(url):
        """Return a list with the translation to this url into each other scheme."""
        other_scheme_urls = []
        match = DocumentationSpider.match_capture_any_scheme.match(url)
        assert match
        if not (match and match.group(1) and match.group(2)):
            raise ValueError(
                "Must have a match and split the url into the scheme and the rest. url: " + url)

        previous_scheme = match.group(1)
        url_with_no_scheme = match.group(2)

        for scheme in DocumentationSpider.every_schemes:
            if scheme != previous_scheme:
                other_scheme_urls.append(scheme + url_with_no_scheme)
        return other_scheme_urls

    def __init__(self, config, meilisearch_helper, strategy, *args, **kwargs):
        # Scrapy config
        self.name = config.index_uid
        self.allowed_domains = config.allowed_domains
        self.start_urls_full = config.start_urls
        self.start_urls = [start_url['url'] for start_url in config.start_urls]
        # We need to ensure that the stop urls are scheme agnostic too if it represents URL
        self.stop_urls = [DocumentationSpider.to_any_scheme(stop_url) for
                          stop_url in config.stop_urls]
        self.meilisearch_helper = meilisearch_helper
        self.strategy = strategy
        self.js_render = config.js_render
        self.js_wait = config.js_wait
        self.scrape_start_urls = config.scrape_start_urls
        self.remove_get_params = config.remove_get_params
        self.strict_redirect = config.strict_redirect
        self.nb_hits_max = config.nb_hits_max
        super().__init__(*args, **kwargs)

        # Get rid of scheme consideration
        # Start_urls must stays authentic URL in order to be reached, we build agnostic scheme regex based on those URL
        start_urls_any_scheme = [DocumentationSpider.to_any_scheme(start_url)
                                 for start_url in self.start_urls]
        link_extractor = LxmlLinkExtractor(
            allow=start_urls_any_scheme,
            deny=self.stop_urls,
            tags=('a', 'area', 'iframe'),
            attrs=('href', 'src'),
            canonicalize=(not config.js_render or not config.use_anchors)
        )

        DocumentationSpider.rules = [
            Rule(link_extractor, callback=self.parse_from_start_url,
                 follow=True),
        ]

        # START _init_ part from SitemapSpider
        # We son't want to check anything if we don't even have a sitemap URL
        if config.sitemap_urls:
            # In case we don't have a special documentation regex,
            # we assume that start_urls are there to match a documentation part
            self.sitemap_urls_regexs =\
                config.sitemap_urls_regexs if config.sitemap_urls_regexs else start_urls_any_scheme

            sitemap_rules = []
            if self.sitemap_urls_regexs:
                for regex in self.sitemap_urls_regexs:
                    sitemap_rules.append((regex, 'parse_from_sitemap'))
            else:  # Neither start url nor regex: default, we parse all
                print("Neither start url nor regex: default, we scrap all")
                sitemap_rules = [('.*', 'parse_from_sitemap')]

            self.__init_sitemap_(config.sitemap_urls, sitemap_rules,
                                 config.sitemap_alternate_links)
            self.force_sitemap_urls_crawling = config.force_sitemap_urls_crawling

        # END _init_ part from SitemapSpider
        super()._compile_rules()

    def start_requests(self):
        # We crawl according to the sitemap
        for url in self.sitemap_urls:
            yield Request(url, callback=self._parse_sitemap,
                          meta={
                              "alternative_links": DocumentationSpider.to_other_scheme(
                                  url)
                          },
                          errback=self.errback_alternative_link)
        # Redirection is neither an error (4XX status) nor a success (2XX) if dont_redirect=False, thus we force it

        # We crawl the start URL in order to ensure we didn't miss anything (Even if we used the sitemap)
        for url in self.start_urls:
            yield Request(url,
                          callback=self.parse_from_start_url if self.scrape_start_urls else self.parse,
                          # If we wan't to crawl (default behavior) without scraping, we still need to let the
                          # crawling spider acknowledge the content by parsing it with the built-in method
                          meta={
                              "alternative_links": DocumentationSpider.to_other_scheme(
                                  url)
                          },
                          errback=self.errback_alternative_link)

    def parse(self, response, **kwargs):
        return super()._parse(response, **kwargs)

    def add_records(self, response, from_sitemap):
        records = self.strategy.get_records_from_response(response)
        self.meilisearch_helper.add_records(records, response.url, from_sitemap)

        DocumentationSpider.NB_INDEXED += len(records)

        # Arbitrary limit
        if self.nb_hits_max > 0 and DocumentationSpider.NB_INDEXED > self.nb_hits_max:
            DocumentationSpider.NB_INDEXED = 0
            self.reason_to_stop = f"Too much hits, Docs-Scraper only handle {int(self.nb_hits_max)} records"
            raise ValueError(self.reason_to_stop)

    def parse_from_sitemap(self, response):
        if self.reason_to_stop is not None:
            raise CloseSpider(reason=self.reason_to_stop)

        if (not self.force_sitemap_urls_crawling) and (
                not self.is_rules_compliant(response)):
            print("\033[94m> Ignored from sitemap:\033[0m " + response.url)
        else:
            self.add_records(response, from_sitemap=True)
            # We don't return self.parse(response) in order to avoid crawling those web page

    def parse_from_start_url(self, response):
        if self.reason_to_stop is not None:
            raise CloseSpider(reason=self.reason_to_stop)

        if self.is_rules_compliant(response):
            self.add_records(response, from_sitemap=False)
        else:
            print("\033[94m> Ignored: from start url\033[0m " + response.url)

        return self.parse(response)

    def is_rules_compliant(self, response):

        # Even if the link extract were compliant, we may have been redirected. Hence we check a new time

        # Redirection redirect on a start url
        if not self.scrape_start_urls and (
                response.url in self.start_urls or response.request.url in self.start_urls):
            return False

        for rule in self._rules:
            if not self.strict_redirect:
                if rule.link_extractor._link_allowed(response):
                    continue

                if rule.link_extractor._link_allowed(response.request):
                    response.replace(url=response.request.url)
                    continue
            else:
                if rule.link_extractor._link_allowed(
                        response) and rule.link_extractor._link_allowed(
                    response.request):
                    continue
            return False

        return True

    def errback_alternative_link(self, failure):
        """
        This error callback will launch the same request with the alternative_links if there are some left
        Only for start_urls and sitemap_urls
        """
        if hasattr(failure.value, 'response'):
            if hasattr(failure.value.response, 'status'):
                self.logger.error('Http Status:%s on %s',
                                  failure.value.response.status,
                                  failure.value.response.url)
            else:
                self.logger.error('Failure : %s', failure.value)

        else:
            self.logger.error('Failure without response %s', failure.value)

        if failure.check(HttpError):
            # these exceptions come from HttpError spider middleware
            meta = failure.request.meta
            meta["alternative_fallback"] = True

            if len(meta["alternative_links"]) > 0:
                alternative_link = meta["alternative_links"].pop(0)
                self.logger.error('Alternative link: %s', alternative_link)
                yield failure.request.replace(
                    url=alternative_link,
                    meta=meta
                )

                # Other check available such as DNSLookupError, TimeoutError, TCPTimedOutError)...

    def __init_sitemap_(self, sitemap_urls, custom_sitemap_rules,
                        sitemap_alternate_links):
        """Init method of a SiteMapSpider @Scrapy"""
        self.sitemap_alternate_links = sitemap_alternate_links
        self.sitemap_urls = sitemap_urls
        self.sitemap_rules = custom_sitemap_rules
        self._cbs = []
        for r, c in self.sitemap_rules:
            if isinstance(c, str):
                c = getattr(self, c)
            self._cbs.append((regex(r), c))
        self._follow = [regex(x) for x in self.sitemap_follow]



import builtins
import re
import copy

from urllib.parse import urlparse
from ..js_executor import JsExecutor


class UrlsParser:
    @staticmethod
    def parse(config_start_urls):
        start_urls = []
        for start_url in config_start_urls:
            if isinstance(start_url, str):
                start_url = {'url': start_url}

            start_url['compiled_url'] = re.compile(start_url['url'])

            if "scrape" not in start_url:
                start_url['scrape'] = True

            if "page_rank" not in start_url:
                start_url['page_rank'] = 0

            if "tags" not in start_url:
                start_url['tags'] = []

            if "selectors_key" not in start_url:
                start_url['selectors_key'] = 'default'

            matches = UrlsParser.get_url_variables_name(start_url['url'])

            start_url['url_attributes'] = {}
            for match in matches:
                if len(start_url['url']) > 2 and start_url['url'][-2:] == '?)':
                    print('\033[0;35mWARNING: ' + start_url[
                        'url'] + ' finish by a variable.'
                                 ' The regex probably won\'t work as expected.'
                                 ' Add a \'/\' or \'$\' to make it work properly\033[0m')
                start_url['url_attributes'][match] = None

            # If there is tag(s) we need to generate all possible urls
            if len(matches) > 0:
                values = {}
                for match in matches:
                    if 'variables' in start_url:
                        if match in start_url['variables']:
                            if isinstance(start_url['variables'][match], list):
                                values[match] = start_url['variables'][match]
                            else:
                                if 'url' in start_url['variables'][
                                    match] and 'js' in start_url['variables'][
                                    match]:
                                    executor = JsExecutor()
                                    values[match] = executor.execute(
                                        start_url['variables'][match]['url'],
                                        start_url['variables'][match]['js'])
                                else:
                                    raise builtins.Exception(
                                        "Bad arguments for variables." + match + " for url " +
                                        start_url['url'])
                        else:
                            raise builtins.Exception(
                                "Missing " + match + " in variables" + " for url " +
                                start_url['url'])

                start_urls = UrlsParser.geturls(start_url, matches[0],
                                                matches[1:], values,
                                                start_urls)

            # If there is no tag just keep it like this
            else:
                start_urls.append(start_url)

        return start_urls

    @staticmethod
    def get_url_variables_name(url):
        # Cache it to avoid to compile it several time
        if not hasattr(UrlsParser.get_url_variables_name, 'group_regex'):
            UrlsParser.get_url_variables_name.group_regex = re.compile(
                r'\(\?P<(.+?)>.+?\)')

        return re.findall(UrlsParser.get_url_variables_name.group_regex, url)

    @staticmethod
    def geturls(start_url, current_match, matches, values, start_urls):
        for value in values[current_match]:
            copy_start_url = copy.copy(start_url)
            copy_start_url['original_url'] = copy_start_url['url']
            copy_start_url['url'] = copy_start_url['url'].replace(
                "(?P<" + current_match + ">.*?)", value)
            copy_start_url['compiled_url'] = re.compile(copy_start_url['url'])
            # Fix reference issue
            copy_start_url['url_attributes'] = copy.deepcopy(
                start_url['url_attributes'])
            copy_start_url['url_attributes'][current_match] = value

            if len(matches) == 0:
                start_urls.append(copy_start_url)
            else:
                start_urls = UrlsParser.geturls(copy_start_url, matches[0],
                                                matches[1:], values,
                                                start_urls)

        return start_urls

    @staticmethod
    def get_extra_facets(start_urls):
        extra_facets = []
        for start_url in start_urls:
            for tag in start_url['url_attributes']:
                extra_facets.append(tag)

        extra_facets = set(extra_facets)

        return list(extra_facets)

    @staticmethod
    def build_allowed_domains(start_urls, stop_urls):
        def get_domain(url):
            """ Return domain name from url """
            return urlparse(url).netloc

        # Concatenating both list, being careful that they can be None
        all_urls = [_['url'] if not isinstance(_, str) else _ for _ in
                    start_urls] + stop_urls
        # Getting the list of domains for each of them
        all_domains = [get_domain(_) for _ in all_urls]
        # Removing duplicates
        all_domains_unique = []
        for domain in all_domains:
            if domain in all_domains_unique:
                continue
            all_domains_unique.append(domain)

        return all_domains_unique

    # Check if tags are defined for the current_page or one of the parent page
    @staticmethod
    def get_tags(current_page_url, start_urls):
        if current_page_url is not None:
            for start_url in start_urls:
                if start_url['compiled_url'].match(current_page_url):
                    return start_url['tags']
        return []

    # Check if a page_rank is defined for the current_page or one of the parent page
    @staticmethod
    def get_page_rank(current_page_url, start_urls):
        if current_page_url is not None:
            for start_url in start_urls:
                if start_url['compiled_url'].match(current_page_url):
                    return int(start_url['page_rank'])
        return 0

    @staticmethod
    def get_extra_attributes(current_page_url, start_urls):
        if current_page_url is not None:
            for start_url in start_urls:
                if start_url['compiled_url'].match(current_page_url):
                    if 'extra_attributes' in start_url:
                        return start_url['extra_attributes']
        return {}

    @staticmethod
    def get_url_variables(current_page_url, start_urls):
        for start_url in start_urls:
            compiled_url = start_url['compiled_url']
            result = re.search(compiled_url, current_page_url)

            if result:
                for attr in start_url['url_attributes']:
                    value = start_url['url_attributes'][attr]
                    if value is not None:
                        current_page_url = current_page_url.replace(value, '')
                        yield attr, value, current_page_url



"""
Default Strategy
"""

import os

from lxml.etree import XPath
from .abstract_strategy import AbstractStrategy
from .anchor import Anchor
from .hierarchy import Hierarchy
from ..config.urls_parser import UrlsParser
from ..helpers import to_json
import json
import hashlib
import sys


class DefaultStrategy(AbstractStrategy):
    """
    DefaultStrategy
    """
    dom = None

    def __init__(self, config):
        super().__init__(config)
        self.levels = ['lvl0', 'lvl1', 'lvl2', 'lvl3', 'lvl4', 'lvl5', 'lvl6']
        self.global_content = {}
        self.page_rank = {}

    def select(self, path):
        """Select an element in the current DOM using specified CSS selector"""
        return XPath(path)(self.dom) if len(path) > 0 else []

    def get_records_from_response(self, response):
        """
        Main method called from the DocumentationSpider. Will be passed the HTTP
        response and will return a list of all records"""

        if self._body_contains_stop_content(response):
            return []

        self.dom = self.get_dom(response)
        self.dom = self.remove_from_dom(self.dom,
                                        self.config.selectors_exclude)

        records = self.get_records_from_dom(response.url)

        return records

    def _update_hierarchy_with_global_content(self, hierarchy,
                                              current_level_int):
        for index in range(0, current_level_int + 1):
            if f'lvl{index}' in self.global_content:
                hierarchy[f'lvl{index}'] = self.global_content[
                    f'lvl{index}']

        return hierarchy

    def _update_record_with_global_content(self, record, levels):
        for level in levels:
            if 'lvl' not in level and level not in ['content', 'text']:
                record[level] = self.global_content[level]

        return record

    def get_records_from_dom(self, current_page_url=None):

        if self.dom is None:
            sys.exit('DefaultStrategy.dom is not defined')

        # Reset it to be able to have a clean instance when testing
        self.global_content = {}

        selectors = self.get_selectors_set(current_page_url)
        levels = self._get_used_levels(selectors)

        # We get a big selector that matches all relevant nodes, in order
        # But we also keep a list of all matches for each individual level
        nodes_per_level = self._get_nodes_per_level(selectors, levels)
        nodes = self._get_all_matching_nodes(levels, selectors)
        self._get_nodes_per_global_level(selectors, levels)

        # We keep the current hierarchy and anchor state between loops
        previous_hierarchy = self._generate_empty_hierarchy()
        anchors = self._generate_empty_hierarchy()

        records = []

        for position, node in enumerate(nodes):
            current_level = self._get_level_of_node(node, nodes_per_level,
                                                    levels)

            # We copy the previous hierarchy, We override the current level, And set all levels after it to None
            hierarchy = previous_hierarchy.copy()

            # Update the hierarchy for each new header
            current_level_int = levels.index(current_level)

            if current_level != 'content':
                hierarchy[current_level] = self._get_text_content_for_level(
                    node, current_level, selectors)
                anchors[current_level] = Anchor.get_anchor(node)

                for index in range(current_level_int + 1, 7):
                    hierarchy[f'lvl{index}'] = None
                    anchors[f'lvl{index}'] = None
                previous_hierarchy = hierarchy

                if self.config.only_content_level:
                    continue

            if current_level_int < self.get_min_indexed_level_for_url(
                    current_page_url):
                continue

            hierarchy = self._update_hierarchy_with_global_content(hierarchy,
                                                                   current_level_int)

            # We only save content for the 'text' matches
            content = None if current_level != 'content' else self.get_text(
                node, self.get_strip_chars(current_level, selectors))

            if (
                    content is None or content == "") and current_level == 'content':
                continue

            hierarchy, content = self._handle_default_values(hierarchy,
                                                             content,
                                                             selectors,
                                                             self.levels)

            # remove trailing ' #' in headers
            for k, v in hierarchy.items():
                if v:
                    if v.endswith(' #'):
                        hierarchy[k] = v[:-2]

            # noinspection PyDictCreation
            record = {
                'anchor': self._get_closest_anchor(anchors),
                'content': content,
                'hierarchy': hierarchy,
                'hierarchy_radio': Hierarchy.get_hierarchy_radio(hierarchy,
                                                                 current_level,
                                                                 levels),
                'type': current_level,
                'tags': UrlsParser.get_tags(current_page_url,
                                            self.config.start_urls),
                'weight': {
                    'page_rank': UrlsParser.get_page_rank(current_page_url,
                                                          self.config.start_urls),
                    'level': self.get_level_weight(current_level),
                    'position': position
                },
                'url': current_page_url,
                'url_without_variables': current_page_url
            }

            extra_attributes = UrlsParser.get_extra_attributes(
                current_page_url, self.config.start_urls)

            for key in list(extra_attributes.keys()):
                record[key] = extra_attributes[key]

            record['hierarchy_camel'] = record['hierarchy'],
            record['hierarchy_radio_camel'] = record['hierarchy_radio']
            record['content_camel'] = record['content']

            self._update_record_with_global_content(record, selectors)

            # get meta data
            for meta_node in self.select('//meta'):
                name = meta_node.get('name')
                content = meta_node.get('content')
                if name and name.startswith('docs-scraper:') and content:
                    name = name.replace('docs-scraper:', '')
                    jsonized = to_json(content)
                    if jsonized:
                        record[name] = jsonized
                    else:
                        record[name] = content

                    if name == "version":
                        record[name] = str(record[name])

            if current_page_url is not None:
                # Add variables to the record
                for attr, value, url_without_variables in UrlsParser.get_url_variables(
                        current_page_url, self.config.start_urls):
                    record['url_without_variables'] = url_without_variables
                    record[attr] = value

                record['url_without_anchor'] = record['url']
                record['url'] = self._get_url_with_anchor(record['url'],
                                                          record['anchor'])
                record['url_without_variables'] = self._get_url_with_anchor(
                    record['url_without_variables'], record['anchor'])
                record['no_variables'] = record['url'] == record[
                    'url_without_variables']

            # Define our own ObjectID to enable proper analytics
            hierarchy_to_hash = {lvl: x for lvl, x in hierarchy.items() if
                                 x is not None}
            raw_hash = hashlib.sha1(json.dumps(
                {'hierarchy_to_hash': hierarchy_to_hash,
                 'url': record['url'],
                 'position': position}, sort_keys=True).encode('utf-8'))
            digest_hash = raw_hash.hexdigest()
            record['objectID'] = digest_hash

            # Replace the localhost and the port with the desired CNAME
            # Example:
            # http://localhost:1314/index.html#pyansys-actions
            cname = os.getenv("DOCUMENTATION_CNAME")
            if cname is not None:
                http, _, port_and_page = record["url"].split(":")
                port = os.getenv("DOCUMENTATION_PORT")
                path = port_and_page[len(port):]
                record['url'] = f"{http}s://{cname}{path}"

            records.append(record)

        return records

    def _get_text_content_for_level(self, node, current_level, selectors):
        if 'attributes' in selectors[current_level]:
            attributes = {}
            for attribute_name in list(selectors[current_level][
                                           'attributes'].keys()):
                matching_nodes = node.xpath(
                    selectors[current_level]['attributes'][attribute_name][
                        'selector'])
                attributes[attribute_name] = self.get_text_from_nodes(
                    matching_nodes,
                    self.get_strip_chars(attribute_name,
                                         selectors[current_level][
                                             'attributes'])
                )
            return attributes

        if current_level in self.global_content:
            return self.global_content[current_level]

        return self.get_text(node,
                             self.get_strip_chars(current_level, selectors))

    @staticmethod
    def _get_closest_anchor(anchors):
        # Getting the element anchor as the closest one
        for index in list(range(6, -1, -1)):
            potential_anchor = anchors[f'lvl{index}']
            if potential_anchor is None:
                continue
            return potential_anchor

        return None

    @staticmethod
    def _generate_empty_hierarchy():
        return {
            "lvl0": None,
            "lvl1": None,
            "lvl2": None,
            "lvl3": None,
            "lvl4": None,
            "lvl5": None,
            "lvl6": None,
        }

    @staticmethod
    def _get_level_of_node(node, nodes_per_level, levels):
        for level in levels:
            if level in nodes_per_level:  # if it's global we won't have it
                if node in nodes_per_level[level]:
                    return level

        return None

    @staticmethod
    def _handle_default_values(hierarchy, content, selectors, levels):
        # Handle default values
        for level in selectors:
            if level in levels:
                if hierarchy[level] is None and selectors[level][
                    'default_value'] is not None:
                    hierarchy[level] = selectors[level]['default_value']
            elif level == 'content':
                if content is None and selectors[level][
                    'default_value'] is not None:
                    content = selectors['content']['default_value']

        return hierarchy, content

    def _get_nodes_per_global_level(self, selectors, levels):
        for level in list(selectors.keys()):
            level_selector = selectors[level]
            if level not in levels or level_selector['global']:
                matching_dom_nodes = self.select(level_selector['selector'])
                self.global_content[level] = self.get_text_from_nodes(
                    matching_dom_nodes,
                    self.get_strip_chars(level, selectors))

                if self.global_content[level] is None and level_selector[
                    'default_value'] is not None:
                    self.global_content[level] = level_selector[
                        'default_value']

    def _get_nodes_per_level(self, selectors, levels):
        nodes_per_level = {}

        for level in levels:
            level_selector = selectors[level]
            matching_dom_nodes = self.select(level_selector['selector'])

            if not level_selector['global']:
                nodes_per_level[level] = matching_dom_nodes

        return nodes_per_level

    def _get_all_matching_nodes(self, levels, selectors):
        return self.select(
            " | ".join(self._get_selector_all(levels, selectors)))

    @staticmethod
    def _get_selector_all(levels, selectors):
        selector_all = []
        for level in levels:
            level_selector = selectors[level]

            if len(level_selector['selector']) > 0 and not level_selector[
                'global']:
                selector_all.append(level_selector['selector'])

        return selector_all

    def _get_used_levels(self, selectors):
        levels = list(self.levels)

        used_levels = []

        for level in levels:
            if level not in selectors:
                break
            used_levels.append(level)

        if 'content' in selectors:
            used_levels.append('content')

        return used_levels

    def _body_contains_stop_content(self, response):
        if len(self.config.stop_content) > 0:
            body = self.get_body(response)

            for stop_content in self.config.stop_content:
                if stop_content in body:
                    return True

        return False

    def _get_url_with_anchor(self, current_page_url, anchor):
        if (
                not self.config.js_render or not self.config.use_anchors) and anchor is not None:
            return current_page_url + '#' + anchor

        return current_page_url



"""
CustomDownloaderMiddleware
"""

import time

from scrapy.http import HtmlResponse
from urllib.parse import urlparse, unquote_plus


class CustomDownloaderMiddleware:
    driver = None

    def __init__(self):
        self.seen = {}  # TODO could be removed
        self.driver = CustomDownloaderMiddleware.driver

    def process_request(self, request, spider):
        if not spider.js_render:
            return None

        if spider.remove_get_params:
            o = urlparse(request.url)
            url_without_params = o.scheme + "://" + o.netloc + o.path
            request = request.replace(url=url_without_params)

        if request.url in self.seen:
            return None

        self.seen[request.url] = True

        print("Getting " + request.url + " from selenium")

        self.driver.get(unquote_plus(
            request.url))  # Decode url otherwise firefox is not happy. Ex /#%21/ => /#!/%21
        time.sleep(spider.js_wait)
        body = self.driver.page_source.encode('utf-8')
        url = self.driver.current_url

        return HtmlResponse(
            url=url,
            body=body,
            encoding='utf8'
        )

    def process_response(self, request, response, spider):
        # Since scrappy use start_urls and stop_urls before creating the request
        # If the url get redirected then this url gets crawled even if it's not allowed to
        # So we check if the final url is allowed

        if spider.remove_get_params:
            o = urlparse(response.url)
            url_without_params = o.scheme + "://" + o.netloc + o.path
            response = response.replace(url=url_without_params)

        if response.url == request.url + '#':
            response = response.replace(url=request.url)

        return response



import json
import time


class JsExecutor:
    driver = None

    def __init__(self):
        self.driver = JsExecutor.driver

    # TODO: find out why JsExecutor.driver couldn't be used
    def execute(self, url, js):
        self.driver.get(url)
        time.sleep(5)

        result = self.driver.execute_script(js)

        try:
            parsed_result = json.loads(result)
            return parsed_result
        except ValueError as value_error:
            raise ValueError('CONFIG is not a valid JSON') from value_error



from scrapy.dupefilters import RFPDupeFilter
from scrapy.utils.job import job_dir
from w3lib.url import canonicalize_url
from scrapy.utils.python import to_bytes
import weakref
import hashlib
import os
import re

_fingerprint_cache = weakref.WeakKeyDictionary()


class CustomDupeFilter(RFPDupeFilter):
    def request_fingerprint(self, request, remove_scheme=None):
        return self.custom_request_fingerprint(request,
                                               remove_scheme=remove_scheme)

    def custom_request_fingerprint(self, request, include_headers=None,
                                   remove_scheme=None):
        """
        Overridden given that some URL can have a wrong encoding (when it is comes from selenium driver) changes:
        encode.('utf-8) & in order to be no scheme compliant
        """

        # If use_anchors, anchors in URL matters since each anchor
        # define a different webpage and content (special js_rendering)
        url_for_finger_print = canonicalize_url(
            request.url) if not self.use_anchors else request.url
        url_for_hash = url_for_finger_print.encode('utf-8')

        # scheme agnosticism
        if remove_scheme:
            match_capture_any_scheme = r'(https?)(.*)'
            url_for_hash = re.sub(match_capture_any_scheme, r"\2",
                                  url_for_finger_print)

        if include_headers:
            include_headers = tuple(to_bytes(h.lower())
                                    for h in sorted(include_headers))
        cache = _fingerprint_cache.setdefault(request, {})

        if include_headers not in cache or not remove_scheme:
            # Since it is called from the same function, wee need to ensure
            # we compute the fingerprint which take into account the scheme.
            # Avoid caching
            fp = hashlib.sha1()
            fp.update(to_bytes(request.method.encode('utf-8')))
            fp.update(to_bytes(url_for_hash))
            fp.update(request.body or ''.encode('utf-8'))
            if include_headers:
                for hdr in include_headers:
                    if hdr in request.headers:
                        fp.update(hdr)
                        for v in request.headers.getlist(hdr):
                            fp.update(v)
            cache[include_headers] = fp.hexdigest()
        return cache[include_headers]

    def __init__(self, path=None, debug=False, use_anchors=False):
        super().__init__(path=path, debug=debug)
        # Spread config bool
        self.use_anchors = use_anchors
        self.fingerprints_with_scheme = set()  # This set will not be scheme agnostic

    # Overridden method in order to add the use_anchors attribute
    @classmethod
    def from_settings(cls, settings):
        debug = settings.getbool('DUPEFILTER_DEBUG')
        use_anchors = settings.getbool('DUPEFILTER_USE_ANCHORS')
        return cls(job_dir(settings), debug, use_anchors)

    def request_seen(self, request):
        """
        Overridden given that we have to handle the redirection case :
        - redirection could be into the same page with another scheme => need the fingerprint
          to take into account the scheme
        - avoid loop in redirection by keeping a dedicated track of it
        """

        fp = self.request_fingerprint(request, remove_scheme=True)
        fp_with_scheme = self.request_fingerprint(request, remove_scheme=False)
        # Request from a redirection which is followed byt he RedirectionMiddleware
        isRedirected = request.meta.get('redirect_times') and request.meta.get(
            'redirect_times') > 0
        isFallback = request.meta.get("alternative_fallback")

        if isRedirected or isFallback:
            if fp_with_scheme in self.fingerprints_with_scheme:
                return True
            self.fingerprints_with_scheme.add(fp_with_scheme)
            # We don't want go back onto the visited web page, especially when it isn't from a redirection
            if fp not in self.fingerprints:
                self.register_fingerprint(fp)
        else:
            if fp in self.fingerprints:
                return True
            self.register_fingerprint(fp)
            self.fingerprints_with_scheme.add(fp_with_scheme)

    def register_fingerprint(self, fp):
        self.fingerprints.add(fp)
        if self.file:
            self.file.write(fp + os.linesep)



"""MeiliSearchHelper
Wrapper on top of the Meilisearch API client"""

import meilisearch
from builtins import range

from .config.version import qualified_version

def remove_bad_encoding(value):
    return value.replace('&#x27;', "'")

def clean_one_field(value):
    if isinstance(value, bool):
        return str(value)
    elif isinstance(value, str):
        return remove_bad_encoding(value)
    return value

def clean_dict(record):
    for key, value in record.items():
        if isinstance(value, dict):
            record[key] = clean_dict(value)
        else:
            record[key] = clean_one_field(value)
    return record

def parse_record(record):
    new_weight = {}
    for k, v in record['weight'].items():
        new_weight[k] = v
    new_hierarchy = {}
    for k, v in record['hierarchy'].items():
        new_hierarchy['hierarchy_' + k] = v
    new_hierarchy_radio = {}
    for k, v in record['hierarchy_radio'].items():
        key = 'hierarchy_radio_' + k
        new_hierarchy_radio = {**{key: v}, **new_hierarchy_radio}
    del record['weight']
    del record['hierarchy']
    del record['hierarchy_radio']
    del record['hierarchy_camel']
    del record['hierarchy_radio_camel']
    del record['content_camel']
    return {**record, **new_weight, **new_hierarchy, **new_hierarchy_radio}

class MeiliSearchHelper:
    """MeiliSearchHelper"""

    # Cf the end of this file to understand these settings
    SETTINGS = {
        'rankingRules': [
            'words',
            'typo',
            'attribute',
            'proximity',
            'exactness',
            'page_rank:desc',
            'level:desc',
            'position:asc'
        ],
        'distinctAttribute': 'url',
        'searchableAttributes': [
            'hierarchy_radio_lvl0',
            'hierarchy_radio_lvl1',
            'hierarchy_radio_lvl2',
            'hierarchy_radio_lvl3',
            'hierarchy_radio_lvl4',
            'hierarchy_radio_lvl5',
            'hierarchy_lvl0',
            'hierarchy_lvl1',
            'hierarchy_lvl2',
            'hierarchy_lvl3',
            'hierarchy_lvl4',
            'hierarchy_lvl5',
            'hierarchy_lvl6',
            'content',
            'objectID',
            'page_rank',
            'level',
            'position'
        ],
        'displayedAttributes': [
            'hierarchy_radio_lvl0',
            'hierarchy_radio_lvl1',
            'hierarchy_radio_lvl2',
            'hierarchy_radio_lvl3',
            'hierarchy_radio_lvl4',
            'hierarchy_radio_lvl5',
            'hierarchy_lvl0',
            'hierarchy_lvl1',
            'hierarchy_lvl2',
            'hierarchy_lvl3',
            'hierarchy_lvl4',
            'hierarchy_lvl5',
            'hierarchy_lvl6',
            'anchor',
            'url',
            'content',
            'objectID'
        ]
    }

    def __init__(self, host_url, api_key, index_uid, custom_settings):
        self.meilisearch_client = meilisearch.Client(host_url, api_key, client_agents=(qualified_version(),))
        self.meilisearch_index = self.meilisearch_client.index(index_uid)
        self.delete_index()
        self.add_settings(MeiliSearchHelper.SETTINGS, custom_settings)

    def add_settings(self, default_settings, custom_settings):
        settings = {**default_settings, **custom_settings}
        self.meilisearch_index.update_settings(settings)

    def delete_index(self):
        self.meilisearch_index.delete()

    def add_records(self, records, url, from_sitemap):
        """Add new records to the index"""

        record_count = len(records)
        for i in range(0, record_count, 50):
            parsed_records = list(map(parse_record, records[i:i + 50]))
            cleaned_records = list(map(clean_dict, parsed_records))
            self.meilisearch_index.add_documents(cleaned_records)

        color = "96" if from_sitemap else "94"

        print(
            f'\033[{color}m> Docs-Scraper: \033[0m{url}\033[93m {record_count} records\033[0m)')

# Algolia's settings:
    # {"minWordSizefor1Typo"=>3,
    # "minWordSizefor2Typos"=>7,
    # "hitsPerPage"=>20,
    # "maxValuesPerFacet"=>100,
    # "minProximity"=>1,
    # "version"=>2,
    # "attributesToIndex"=>
    # ["unordered(hierarchy_radio_camel.lvl0)",
    # "unordered(hierarchy_radio.lvl0)",
    # "unordered(hierarchy_radio_camel.lvl1)",
    # "unordered(hierarchy_radio.lvl1)",
    # "unordered(hierarchy_radio_camel.lvl2)",
    # "unordered(hierarchy_radio.lvl2)",
    # "unordered(hierarchy_radio_camel.lvl3)",
    # "unordered(hierarchy_radio.lvl3)",
    # "unordered(hierarchy_radio_camel.lvl4)",
    # "unordered(hierarchy_radio.lvl4)",
    # "unordered(hierarchy_radio_camel.lvl5)",
    # "unordered(hierarchy_radio.lvl5)",
    # "unordered(hierarchy_camel.lvl0)",
    # "unordered(hierarchy.lvl0)",
    # "unordered(hierarchy_camel.lvl1)",
    # "unordered(hierarchy.lvl1)",
    # "unordered(hierarchy_camel.lvl2)",
    # "unordered(hierarchy.lvl2)",
    # "unordered(hierarchy_camel.lvl3)",
    # "unordered(hierarchy.lvl3)",
    # "unordered(hierarchy_camel.lvl4)",
    # "unordered(hierarchy.lvl4)",
    # "unordered(hierarchy_camel.lvl5)",
    # "unordered(hierarchy.lvl5)",
    # "content"],
    # "numericAttributesToIndex"=>nil,
    # "attributesToRetrieve"=>["hierarchy", "content", "anchor", "url"],
    # "allowTyposOnNumericTokens"=>false,
    # "ignorePlurals"=>true,
    # "camelCaseAttributes"=>["hierarchy", "hierarchy_radio", "content"],
    # "advancedSyntax"=>true,
    # "attributeCriteriaComputedByMinProximity"=>true,
    # "distinct"=>true,
    # "unretrievableAttributes"=>nil,
    # "optionalWords"=>nil,
    # "userData"=>{"crawling_issue"=>false},
    # "attributesForFaceting"=>["lang"],
    # "attributesToSnippet"=>["content:10"],
    # "attributesToHighlight"=>["hierarchy", "hierarchy_camel", "content"],
    # "paginationLimitedTo"=>1000,
    # "attributeForDistinct"=>"url",
    # "exactOnSingleWordQuery"=>"attribute",
    # "ranking"=>
    # ["words", "filters", "typo", "attribute", "proximity", "exact", "custom"],
    # "customRanking"=>
    # ["desc(weight.page_rank)", "desc(weight.level)", "asc(weight.position)"],
    # "separatorsToIndex"=>"",
    # "removeWordsIfNoResults"=>"allOptional",
    # "queryType"=>"prefixLast",
    # "highlightPreTag"=>"<span class=\"algolia-docsearch-suggestion--highlight\">",
    # "highlightPostTag"=>"</span>",
    # "snippetEllipsisText"=>"",
    # "alternativesAsExact"=>["ignorePlurals", "singleWordSynonym"]}



from lxml.cssselect import CSSSelector
import lxml
import re
import json
import html

"""
Abstract Strategy
"""


class AbstractStrategy:
    """
    Abstract Strategy
    """
    config = None

    def __init__(self, config):
        self.config = config
        AbstractStrategy.keep_tags = config.keep_tags

    @staticmethod
    def pprint(data):
        """Pretty print a JSON-like structure"""
        print(
            json.dumps(data, indent=2, sort_keys=True, separators=(',', ': ')))

    @staticmethod
    def get_body(response):
        try:
            body = response.body.decode(response.encoding)
            return body
        except (UnicodeError, ValueError, AttributeError):
            return response.body

    @staticmethod
    def get_dom(response):
        """Get the DOM representation of the webpage"""
        try:
            body = response.body.decode(response.encoding)
            result = lxml.html.fromstring(body)
        except (UnicodeError, ValueError, lxml.etree.ParserError):
            result = lxml.html.fromstring(response.body)
        return result

    def get_strip_chars(self, level, selectors):
        if selectors[level]['strip_chars'] is None:
            return self.config.strip_chars
        return selectors[level]['strip_chars']

    def get_selectors_set_key(self, url):
        selectors_key = 'default'

        if url is not None:
            for start_url in self.config.start_urls:
                if re.search(start_url['compiled_url'], url) is not None:
                    selectors_key = start_url['selectors_key']
                    break

        return selectors_key

    def get_selectors_set(self, url):
        selectors_key = self.get_selectors_set_key(url)

        if selectors_key not in self.config.selectors:
            return self.config.selectors['default']

        return self.config.selectors[selectors_key]

    def get_min_indexed_level_for_url(self, url):
        selectors_key = self.get_selectors_set_key(url)

        if selectors_key not in self.config.min_indexed_level:
            return self.config.min_indexed_level['default']

        return self.config.min_indexed_level[selectors_key]

    @staticmethod
    def itertext(node):
        tag = node.tag
        if not isinstance(tag, str) and tag is not None:
            return

        if node.text:
            if node.tag not in AbstractStrategy.keep_tags:
                yield node.text
            else:
                yield '<' + node.tag + '>' + node.text + '</' + node.tag + '>'
        for e in node:
            for s in AbstractStrategy.itertext(e):
                yield s
            if e.tail:
                yield e.tail

    @staticmethod
    def escape(text):
        text = html.escape(text)

        for tag in AbstractStrategy.keep_tags:
            opening_tag = "<" + tag + ">"
            closing_tag = "</" + tag + ">"
            text = text.replace(html.escape(opening_tag), opening_tag)
            text = text.replace(html.escape(closing_tag), closing_tag)

        text = text.replace('&amp;', '&')

        return text

    @staticmethod
    def get_text(element, strip_chars=None):
        """Return the text content of a DOM node"""
        text = element

        # Do not call text_content if not needed (Ex. xpath selector with text() doesn't return a node but a string)
        if not isinstance(text, str):
            text = ""
            for s in AbstractStrategy.itertext(element):
                text = text + " " + s

        # We call strip a first time for space, tab, newline, return and formfeed
        text = text.strip()
        # then one more time if with custom strip_chars if there is some
        if strip_chars is not None:
            text = text.strip(strip_chars)

        if len(text) == 0:
            return None

        return AbstractStrategy.escape(text)

    @staticmethod
    def get_text_from_nodes(elements, strip_chars=None):
        """
        Return the text content of a set of DOM nodes.
        elements can contain either an array of nodes or a custom data return from xpath
        (Example: integer in the case of a count)
        """

        if not isinstance(elements, list):
            return elements

        if len(elements) == 0:
            return None

        text = ' '.join(
            [AbstractStrategy.get_text(element, strip_chars) for element in
             elements
             if AbstractStrategy.get_text(element, strip_chars) is not None])

        if len(text) == 0:
            return None

        return AbstractStrategy.escape(text)

    @staticmethod
    def remove_from_dom(dom, exclude_selectors):
        """Remove any elements matching the selector from the DOM"""
        for selector in exclude_selectors:
            matches = CSSSelector(selector)(dom)
            if len(matches) == 0:
                continue

            for match in matches:
                match.drop_tree()

        return dom

    @staticmethod
    def elements_are_equals(el1, el2):
        """Checks if two elements are actually the same"""
        return el1.getroottree().getpath(el1) == el2.getroottree().getpath(el2)

    @staticmethod
    def get_level_weight(level):
        """
        Returns a numeric weight based on the level of the selector

        >>> self.get_level_weight('lvl0') = 100
        >>> self.get_level_weight('lvl3') = 70
        >>> self.get_level_weight('content') = 0
        """
        matches = re.match(r'lvl([0-9]*)', level)
        if matches:
            return 100 - int(matches.group(1)) * 10
        return 0



# coding: utf-8

EXIT_CODE_WRONG_CONFIG = 5
"""
Load the config json file.
"""

from collections import OrderedDict
import json
import os
import copy

from .version import qualified_version
from .config_validator import ConfigValidator
from .urls_parser import UrlsParser
from .selectors_parser import SelectorsParser
from .browser_handler import BrowserHandler


class ConfigLoader:
    """
    ConfigLoader
    """
    # We define them here so the linters/autocomplete know what to expect
    allowed_domains = None
    api_key = None
    app_id = None
    custom_settings = {}
    extra_records = []
    index_uid = None
    index_uid_tmp = None
    js_wait = 0
    js_render = False
    keep_tags = []
    min_indexed_level = 0
    remove_get_params = False
    scrap_start_urls = True
    scrape_start_urls = True
    selectors = None
    selectors_exclude = []
    start_urls = []
    stop_urls = []
    stop_content = []
    strategy = 'default'
    strict_redirect = True
    strip_chars = ".,;:§¶"
    use_anchors = False
    user_agent = qualified_version()
    only_content_level = False
    query_rules = []

    # data storage, starting here attribute are not config params
    config_file = None
    config_content = None
    config_original_content = None

    driver = None

    sitemap_alternate_links = False
    sitemap_urls = []
    sitemap_urls_regexs = []
    force_sitemap_urls_crawling = False

    nb_hits_max = 6000000

    def __init__(self, config):
        data = self._load_config(config)

        # Fill self from config
        for key, value in list(data.items()):
            setattr(self, key, value)

        # Start browser if needed
        self.driver = BrowserHandler.init(self.config_original_content,
                                          self.js_render,
                                          self.user_agent)

        # Validate
        ConfigValidator(self).validate()

        # Modify
        self._parse()

        # Stop browser if needed
        if not self.js_render:
            self.driver = BrowserHandler.destroy(self.driver)

        # BC new correct naming
        self.scrape_start_urls = self.scrap_start_urls if not self.scrap_start_urls else self.scrape_start_urls

    def _load_config(self, config):
        if os.path.isfile(config):
            self.config_file = config
            with open(self.config_file, mode='r', encoding='utf-8') as f:
                config = f.read()

        try:
            self.config_original_content = config
            data = json.loads(config, object_pairs_hook=OrderedDict)
            self.config_content = copy.deepcopy(data)

            return data
        except ValueError as value_error:
            raise ValueError('CONFIG is not a valid JSON') from value_error

    def _parse(self):
        # Parse Env
        self.app_id = os.environ.get('MEILISEARCH_HOST_URL', None)
        self.api_key = os.environ.get('MEILISEARCH_API_KEY', None)

        if self.index_uid_tmp is None:
            self.index_uid_tmp = os.environ.get('index_uid_TMP', self.index_uid + '_tmp')

        # Parse config
        self.selectors = SelectorsParser().parse(self.selectors)
        self.min_indexed_level = SelectorsParser().parse_min_indexed_level(
            self.min_indexed_level)
        self.start_urls = UrlsParser.parse(self.start_urls)

        # Build default allowed_domains from start_urls and stop_urls
        if self.allowed_domains is None:
            self.allowed_domains = UrlsParser.build_allowed_domains(
                self.start_urls, self.stop_urls)

    def get_extra_facets(self):
        return UrlsParser.get_extra_facets(self.start_urls)





class Anchor:
    def __init__(self):
        pass

    @staticmethod
    def _get_anchor_string_from_element(element):
        return element.get('name', element.get('id'))

    @staticmethod
    def get_anchor(element):
        """
        Return a possible anchor for that element.
        Looks for name and id, and if not found will look in children
        """
        if isinstance(element, str):
            return None

        # Check the name or id on the element
        anchor = Anchor._get_anchor_string_from_element(element)

        if anchor is not None and anchor != '':
            return anchor

        # Check on child
        children = element.cssselect('[name],[id]')
        if len(children) > 0:
            return Anchor._get_anchor_string_from_element(children[-1])

        el = element

        while el is not None:
            # go back
            while el.getprevious() is not None:
                el = el.getprevious()

                if el is not None:
                    anchor = Anchor._get_anchor_string_from_element(el)

                    if anchor is not None:
                        return anchor

            # check last previous
            if el is not None:
                anchor = Anchor._get_anchor_string_from_element(el)

                if anchor is not None:
                    return anchor

            # go up
            el = el.getparent()

            if el is not None:
                anchor = Anchor._get_anchor_string_from_element(el)

                # check parent
                if anchor is not None:
                    return anchor

        # No more parent, we have no anchor
        return None



import re

from selenium import webdriver
from selenium.webdriver.chrome.options import Options

from ..custom_downloader_middleware import CustomDownloaderMiddleware
from ..js_executor import JsExecutor


class BrowserHandler:
    @staticmethod
    def conf_need_browser(config_original_content, js_render):
        group_regex = re.compile(r'\(\?P<(.+?)>.+?\)')
        results = re.findall(group_regex, config_original_content)

        return len(results) > 0 or js_render

    @staticmethod
    def init(config_original_content, js_render, user_agent):
        driver = None

        if BrowserHandler.conf_need_browser(config_original_content,
                                            js_render):
            chrome_options = Options()
            chrome_options.add_argument('--no-sandbox')
            chrome_options.add_argument('--headless')
            chrome_options.add_argument(f'user-agent={user_agent}')

            driver = webdriver.Chrome(
                options=chrome_options)
            CustomDownloaderMiddleware.driver = driver
            JsExecutor.driver = driver
        return driver

    @staticmethod
    def destroy(driver):
        # Start browser if needed
        if driver is not None:
            driver.quit()
            driver = None

        return driver



import json
import unicodedata
from builtins import input
from cssselect import HTMLTranslator


def confirm(message="Confirm"):
    prompt = message + ' [y/n]:\n'

    while True:
        ans = input(prompt)
        if ans not in ['y', 'Y', 'n', 'N']:
            print('please enter y or n.')
            continue
        if ans == 'y' or ans == 'Y':
            return True
        if ans == 'n' or ans == 'N':
            return False


def css_to_xpath(css):
    return HTMLTranslator().css_to_xpath(css) if len(css) > 0 else ""


def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        pass

    try:
        unicodedata.numeric(s)
        return True
    except (TypeError, ValueError):
        pass
    return False


def to_json(my_json):
    try:
        jsonized = json.loads(my_json)
    except ValueError:
        return None
    return jsonized



"""
Docs-scraper main entry point
"""
import os
import sys
import json
import requests
from requests_iap import IAPAuth
from keycloak.realm import KeycloakRealm

from scrapy.crawler import CrawlerProcess

from .meilisearch_helper import MeiliSearchHelper
from .config.config_loader import ConfigLoader
from .documentation_spider import DocumentationSpider
from .strategies.default_strategy import DefaultStrategy
from .custom_downloader_middleware import CustomDownloaderMiddleware
from .custom_dupefilter import CustomDupeFilter
from .config.browser_handler import BrowserHandler

try:
    # disable boto (S3 download)
    from scrapy import optional_features

    if 'boto' in optional_features:
        optional_features.remove('boto')
except ImportError:
    pass

EXIT_CODE_NO_RECORD = 3


def run_config(config):
    config = ConfigLoader(config)
    CustomDownloaderMiddleware.driver = config.driver
    DocumentationSpider.NB_INDEXED = 0

    strategy = DefaultStrategy(config)

    meilisearch_helper = MeiliSearchHelper(
        config.app_id,
        config.api_key,
        config.index_uid,
        config.custom_settings
    )

    root_module = 'src.' if __name__ == '__main__' else 'scraper.src.'
    DOWNLOADER_MIDDLEWARES_PATH = root_module + 'custom_downloader_middleware.' + CustomDownloaderMiddleware.__name__
    DUPEFILTER_CLASS_PATH = root_module + 'custom_dupefilter.' + CustomDupeFilter.__name__

    headers = {
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
        "Accept-Language": "en",
    }  # Defaults for scrapy https://docs.scrapy.org/en/latest/topics/settings.html#default-request-headers

    if os.getenv("CF_ACCESS_CLIENT_ID") and os.getenv("CF_ACCESS_CLIENT_SECRET"):
        headers.update(
            {
                "CF-Access-Client-Id": os.getenv("CF_ACCESS_CLIENT_ID"),
                "CF-Access-Client-Secret": os.getenv("CF_ACCESS_CLIENT_SECRET"),
            }
        )
    elif os.getenv("IAP_AUTH_CLIENT_ID") and os.getenv("IAP_AUTH_SERVICE_ACCOUNT_JSON"):
        iap_token = IAPAuth(
            client_id=os.getenv("IAP_AUTH_CLIENT_ID"),
            service_account_secret_dict=json.loads(
                os.getenv("IAP_AUTH_SERVICE_ACCOUNT_JSON")
            ),
        )(requests.Request()).headers["Authorization"]
        headers.update({"Authorization": iap_token})
    elif os.getenv("KC_URL") and os.getenv("KC_REALM") and os.getenv("KC_CLIENT_ID") and os.getenv("KC_CLIENT_SECRET"):
        realm = KeycloakRealm(
          server_url=os.getenv("KC_URL"),
          realm_name=os.getenv("KC_REALM"))
        oidc_client = realm.open_id_connect(
          client_id=os.getenv("KC_CLIENT_ID"),
          client_secret=os.getenv("KC_CLIENT_SECRET"))
        token_response = oidc_client.client_credentials()
        token = token_response["access_token"]
        headers.update({"Authorization": 'Bearer ' + token})

    DEFAULT_REQUEST_HEADERS = headers

    process = CrawlerProcess({
        'LOG_ENABLED': '1',
        'LOG_LEVEL': 'ERROR',
        'USER_AGENT': config.user_agent,
        'DOWNLOADER_MIDDLEWARES': {DOWNLOADER_MIDDLEWARES_PATH: 900},
        # Need to be > 600 to be after the redirectMiddleware
        'DUPEFILTER_USE_ANCHORS': config.use_anchors,
        # Use our custom dupefilter in order to be scheme agnostic regarding link provided
        'DUPEFILTER_CLASS': DUPEFILTER_CLASS_PATH,
        'DEFAULT_REQUEST_HEADERS': DEFAULT_REQUEST_HEADERS,
    })

    process.crawl(
        DocumentationSpider,
        config=config,
        meilisearch_helper=meilisearch_helper,
        strategy=strategy
    )

    process.start()
    process.stop()

    # Kill browser if needed
    BrowserHandler.destroy(config.driver)

    if len(config.extra_records) > 0:
        meilisearch_helper.add_records(config.extra_records, "Extra records", False)

    print("")

    if DocumentationSpider.NB_INDEXED > 0:
        # meilisearch_helper.commit_tmp_index()
        print(f'Nb hits: {DocumentationSpider.NB_INDEXED}')
    else:
        print('Crawling issue: nbHits 0 for ' + config.index_uid)
        # meilisearch_helper.report_crawling_issue()
        sys.exit(EXIT_CODE_NO_RECORD)
    print("")


if __name__ == '__main__':
    from os import environ

    run_config(environ['CONFIG'])



from __future__ import annotations

__version__ = "0.12.11"


def qualified_version() -> str:
    """Get the qualified version of this module."""

    return f"Meilisearch DocsScraper (v{__version__})"



"""Main module

Permits:

"""
import os
from scraper.src.index import run_config
import argparse

def main(sys_args=None):
    """ Supports calling gui directly with command line arguments """

    parser = argparse.ArgumentParser(
        description='Scrape documentation.'
    )
    parser.add_argument(
        'config_file', type=str, help='The path to the configuration file.'
    )

    parser.add_argument(
        '--meilisearch-host-url', type=str,
        required=False,
        help='The URL to the meilisearch host',
    )

    parser.add_argument(
        '--meilisearch-api-key', type=str,
        required=False,
        help='The URL to the meilisearch host',
    )

    parser.add_argument(
        '--cname', type=str,
        required=False,
        help='The CNAME of the webpage',
    )

    parser.add_argument(
        '--port', type=str,
        required=False,
        help='The port where local HTML files are served.',
    )

    args = parser.parse_args()

    if args.meilisearch_host_url is not None:
        os.environ['MEILISEARCH_HOST_URL'] = args.meilisearch_host_url
    if args.meilisearch_api_key is not None:
        os.environ['MEILISEARCH_API_KEY'] = args.meilisearch_api_key
    if args.cname is not None:
        os.environ['DOCUMENTATION_CNAME'] = args.cname
    if args.port is not None:
        os.environ['DOCUMENTATION_PORT'] = args.port

    if 'MEILISEARCH_HOST_URL' not in os.environ:
        raise RuntimeError(
            '\n\nMEILISEARCH_HOST_URL is required either the command line argument:'
            '\n\n    --meilisearch-host-url <URL>\n\n'
            'or as the environment variable "MEILISEARCH_HOST_URL"'
        )

    if 'MEILISEARCH_API_KEY' not in os.environ:
        raise RuntimeError(
            '\n\nMEILISEARCH_API_KEY is required either the command line argument:'
            '\n\n    --meilisearch-api-key <URL>\n\n'
            'or as the environment variable "MEILISEARCH_API_KEY"'
        )

    run_config(args.config_file)


if __name__ == '__main__':
    main()
